/*
        newpbio.c
*/
#include "bufrgrib.h"
#include "fort2c.h"

#include "bufrgrib.h"
#include <stdio.h>
#include <string.h>
#ifdef sun4
#include <memory.h>
#endif

#define NAMEBUFFLEN 256
#define MODEBUFFLEN 10
#define MAXFILES 100
#define OFFFILES 1000

#define len3oct(p) ( 256*(256*(int)*((p)) + (int)*((p+1)) ) + (int)*((p+2)) )
#define SMALL 8
#define LARGEBUF 200000
#define GRIB 0x47524942
#define BUFR 0x42554652
#define BUDG 0x42554447
#define TIDE 0x54494445
static FILE* fptable[MAXFILES];

int readprod( char *, char * , int * , 
              int (*read_func)(char *, int , void *), void * );
static int file_read(char *, int , void * stream);
static int prodsize(int , char * , int * , int (*read_func)(), void * );
static int gribsize(char * , int * ,  int (*read_func)(), void * );
static int bufrsize(char * , int * ,  int (*read_func)(), void * );
static int tide_budg_size(char * , int * ,  int (*read_func)(), void * );
static int lentotal(char *, int *, int , int , int , int (*read_func)(), void *) ;

/*************************************************************************
*  FUNCTION:  pbopen - Open file (from FORTRAN)
**************************************************************************
*/
/*void PBOPEN(FILE ** unit, _fcd name, _fcd mode, int * iret, int l1, int l2)*/
void PBOPEN(int * unit, _fcd name, _fcd mode, int * iret) 
/*
* Purpose:      Opens file, return UNIX FILE pointer.
*
* Function returns:     int iret:  -1 = Could not open file.
*                               -2 = Invalid file name.
*                               -3 = Invalid open mode specified
*                                0 = OK.
*
*    Note: l1 and l2 are the lengths of the character strings in 
*          name and mode on SGI.
*/
{
        char *fname;
        char *modes;
        char *p;
        char  flags[4];
        char namebuff[NAMEBUFFLEN], modebuff[MODEBUFFLEN];

        int n;

/* Put the character strings into buffers and ensure that there is a
   null terminator (for SGI case when FORTRAN CHARACTER variable is full
   right to end with characters */
        strncpy( namebuff, name, NAMEBUFFLEN - 1);
        strncpy( modebuff, mode, MODEBUFFLEN - 1);
/*#if defined sgi || defined sun 
        namebuff[l1] = '\0';
        modebuff[l2] = '\0';
#endif*/

        strcpy(flags,"");

        *unit = NULL;
        *iret = 0;

        /* convert fortran to c string : file name */

#if defined sgi || defined sun 
        if(!(fname = fcd2char(namebuff)))
#else
        if(!(fname = fcd2char(name)))
#endif
        {
                *iret = -2;
                return;
        }

        /* strip trailing blanks */

        p  = fname;
        while(*p)
        {
                if(*p == ' ') *p = 0;
                p++;
        }
                
        /* convert fortran to c string : open modes  */
        
#if defined sgi || defined sun 
        if(!(modes = fcd2char(modebuff)))
#else
        if(!(modes = fcd2char(mode)))
#endif
        {
                free(fname);
                *iret = -2;
                return;
        }


        /* build open flags from "modes" */

        p = modes;
        while(*p && ( strlen(flags) < 3 ) )
        {
                switch(*p)
                {
                        case 'a':
                        case 'A':
                                strcat(flags, "a");
                                break;

                        case 'c':
                        case 'C':

                        case 'w':
                        case 'W':
                                strcat(flags, "w");
                                break;

                        case 'r':
                        case 'R':
                                strcat(flags, "r");
                                break;

                        default:
                                *iret = -3;
                                return;

                }
                p++;
        }

        /* if read/write change flags */

        if ( !strcmp(flags,"wr") || !strcmp(flags, "rw") )
                strcpy(flags, "r+w" );

/* look for a free slot on fptable */

        n = 0;
        while (n<MAXFILES)
        {
                if (fptable[n]==0)
                {
                        *unit = n+OFFFILES;
                        break;
                }
                n++;
        }

        if (n==MAXFILES)
        {
                *iret = -1;
                return;
        }

        fptable[n] = fopen(fname, flags );

        if(fptable[n] == NULL)
        {
                perror(fname);
                *iret = -1;
        }



        free(fname);
        free(modes);

}

/*************************************************************************
*  FUNCTION:  pbseek_ - Seek (from FORTRAN)
**************************************************************************
*/
/*void PBSEEK(FILE ** unit, int * offset, int * whence, int * iret)*/
void PBSEEK(int * unit, int * offset, int * whence, int * iret)
/*
*
* Purpose:      Seeks to specified location in file.
*
* Function returns:     int status :   -2 = error in handling file,
*                                       -1 = end-of-file
*                              otherwise,  = byte offset from start of file.
*
*       whence  = 0, from start of file
*               = 1, from current position
*               = 2, from end of file.  
*/
{
        long int my_offset;
        int my_whence;
        my_offset = *offset;
        my_whence = *whence;

        if ( my_whence == 2) my_offset = - abs(my_offset);
                                /* must use negative offset if working
                                   from end-of-file     */

        *iret = ftell(fptable[*unit-OFFFILES]);
        if (*iret==my_offset && my_whence == 0)
                *iret = 0;
        else
        *iret = fseek(fptable[*unit-OFFFILES], my_offset, my_whence);

        if(*iret != 0)
        {
                if ( ! feof(fptable[*unit-OFFFILES]) )
                {
                        *iret = -2;             /* error in file-handling */
                        perror("pbseek");
                }
                else
                        *iret = -1;             /* end-of-file  */

                clearerr(fptable[*unit-OFFFILES]);
                return;
        }

        *iret = ftell(fptable[*unit-OFFFILES]);  /* byte offset from start of file */
 
}
                


 
/*************************************************************************
*  FUNCTION:  pbread_ - Read (from FORTRAN)
**************************************************************************
*/
/*void PBREAD(FILE ** unit, char * buffer, int * nbytes, int * iret)*/
void PBREAD(int * unit, char * buffer, int * nbytes, int * iret)
/*
*
* Purpose:      Reads a block of bytes from a file.
*
* Function returns:     int status :   -2 = error in reading file,
*                                       -1 = end-of-file,
*                               otherwise, = number of bytes read.
*/
{
 
        if ( (*iret = fread(buffer, 1, *nbytes,fptable[*unit-OFFFILES]) ) != *nbytes)
        {
                                                /*      Read problem */
                if ( ! feof(fptable[*unit-OFFFILES]) )
                {
                        *iret = -2;             /*  error in file-handling  */
                        perror("pbread");
                        clearerr(fptable[*unit-OFFFILES]);
                        return;
                }
                else
                {
                        *iret = -1;             /*  end-of-file */
                        clearerr(fptable[*unit-OFFFILES]);
                }
        }
 
}
 

/*************************************************************************
*  FUNCTION:  pbread2_ - Read (from FORTRAN)
**************************************************************************
*/
/*void PBREAD2(FILE ** unit, char * buffer, int * nbytes, int * iret)*/
void PBREAD2(int * unit, char * buffer, int * nbytes, int * iret)
/*
*
* Purpose:      Reads a block of bytes from a file.
*
* Function returns:     int status :   -2 = error in reading file,
*                               otherwise, = number of bytes read.
*/
{

        if ( (*iret = fread(buffer, 1, *nbytes,fptable[*unit-OFFFILES]) ) != *nbytes)
        {
                                                /*      Read problem */
                if ( ! feof(fptable[*unit-OFFFILES]) )
                {
                        *iret = -2;             /*  error in file-handling  */
                        perror("pbread");
                        clearerr(fptable[*unit-OFFFILES]);
                        return;
                }
        }
 
}

/*************************************************************************
*  FUNCTION:  pbwrite_ - Write (from FORTRAN)
**************************************************************************
*/
/*void PBWRITE( FILE ** unit, char * buffer, int * nbytes, int *iret)*/
void PBWRITE( int * unit, char * buffer, int * nbytes, int *iret)
/*
* Purpose:      Writes a block of bytes to a file.
*
* Function returns:     int status : -1 = Could not write to file.
*                                    >=0 = Number of bytes written.
*/
{
        if ( (*iret = fwrite(buffer, 1, *nbytes,fptable[*unit-OFFFILES]) ) != *nbytes)
        {                               /* Problem with write */   
                perror("pbwrite");
                *iret = -1;
        }
    
}



/*************************************************************************
*  FUNCTION:  pbclos_ - clos (from FORTRAN)
**************************************************************************
*/
/*int PBCLOSE( FILE ** unit, int * iret)*/
int PBCLOSE( int * unit, int * iret)
/*
*
* Purpose:      Closes file.
*
* Function returns:     int status : non-0 = error in handling file.
*                                         0 = OK.
*/
{
        *iret = fclose(fptable[*unit-OFFFILES]);

        if(*iret != 0) perror("pbclose");

}

/*
        readprod.c
*/

static int file_read(char * buff, int leng, void * file)
/*  
    buff = buffer to fill,
    leng = size of buff in bytes,
    file = FILE *.

    Returns the number of bytes read.
    On EOF, returns negative value for number of bytes read .
*/
{
int nbytes;

    nbytes = (int) fread( buff, 1, leng, (FILE *) file);

/* If EOF, return negative number of bytes read */
    if ( feof((FILE *) file ) )
    {
        clearerr( (FILE *) file );
        return (-nbytes);
    }

    return nbytes;
}

static int prodsize(int code, char * hold, int * holdsize, 
                     int (*read_func)(), 
                     void * stream)
/*
    Returns size of BUFR, GRIB, BUDG product in bytes.

    hold = buffer holding product,
    leng = size of buffer in bytes,
    read_func = function used to read the product,
    stream = data describing the input stream (eg. FILE *).
*/
{
    *holdsize = 4;
    switch( code ) 
    {
        case BUFR:
            memcpy(hold,"BUFR",4);
            return bufrsize( hold, holdsize, read_func, stream);

        case BUDG:
            memcpy(hold,"BUDG",4);
            return tide_budg_size( hold, holdsize, read_func, stream);

        case GRIB:
            memcpy(hold,"GRIB",4);
            return gribsize( hold, holdsize, read_func, stream);

        case TIDE:
            memcpy(hold,"TIDE",4);
            return tide_budg_size( hold, holdsize, read_func, stream);

        default:
            return 0;
    }
}

int readprod( char * prod_id, char * buffer, int * size, 
              int (*read_func)(char *, int , void *), void * stream)
/*
    Reads a BUFR or GRIB product.

    prod_id = "BUFR", "GRIB", "BUDG", "TIDE" ...
    buffer = buffer to hold product,
    size = on input, size of buffer in bytes,
           on output, number of bytes read.
    read_func = function used to read the product,
    stream = data describing the input stream (eg. FILE *).

    Returns 
            -1 if can't find product  (eg. end-of-file)
            -3 if buffer too small for whole product.
            -4 if buffer too small to get started on product

*/
{
static char * hold = NULL;                      /* buffer used to read product */
int holdsize = SMALL, numread;
int found = 0, num = 0, length, plength;
int original_len = *size;
char p;
unsigned int code=0, wanted = 0;
int status = 0, i;

/* See if user gave a buffer for holding the product */
    if ( buffer == NULL )
    {                                        /* No buffer given, get some .. */
        if ( hold == NULL )                  /* .. but only first time round */
        {
            hold = (char *) malloc( LARGEBUF);
            if ( hold == NULL )
            {
                perror("malloc failed in readnext");
                exit(1);
            }
        }
        original_len = LARGEBUF;
    }
    else                                     /* User buffer given */
        hold = buffer;

/* Make sure the user gave some buffer */
   if (  original_len < holdsize )
    {
        return -4;
    }

/* Look for product identifier */
    if ( prod_id != NULL )
        for ( i = 0; i < 4; i++ )
            wanted = ( (wanted << 8) + *(prod_id+i) ) & 0xFFFFFFFF;

/* Read some bytes from the product */
    do 
    {
        numread = read_func( &p, 1L, stream );
        if ( numread <= 0 ) 
        {
            *size = 0;
            return -1;
        }

        code = ( (code << 8) + p ) & 0xFFFFFFFF;

        if ( prod_id == NULL )
        {
            switch(code)
            {
                case BUDG:
                case BUFR:
                case GRIB:
                case TIDE:
                    found = 1;
            }
        }
        else
        {
            if ( code == wanted )
                found = 1;
        }

    } while ( ! found );

/* Pick up the product length */
    plength = prodsize( code, hold, &holdsize, read_func, stream);

/* Pick up the rest of the product, or as much as possible */

/* Handle case that the user gave a NULL buffer and zero length */
    if( *size == 0 )
    {                                   /* move to end of product */
        status = fseek( (FILE *)stream, plength - holdsize, SEEK_CUR );
        if (status)
        {
            perror("fseek error in readprod");
            exit(1);
        }
        *size = plength;                /* give back actual product length */
        return -3;                      /* indicate buffer too small */
    }

/* Otherwise read product into the buffer */
    length = (original_len > plength) ? plength : original_len ;
    if ( length > holdsize )
    {
        numread = read_func( hold+holdsize, length - holdsize , stream );
        if ( original_len < plength )
        {
            *size = plength;               /* give back actual product length */
            status = fseek( (FILE *)stream, plength - original_len, SEEK_CUR );
                                           /* move to end of product */
            if (status)
            {
                perror("fseek error in readprod");
                exit(1);
            }
            return -3;                      /* indicate buffer too small */
        }
    }

    *size = plength;                       /* give back actual product length */
    return *size;

}

#define BIT1 0x80
#define BIT2 0x40
#define BIT7 0x02
#define BIT8 0x01


static int gribsize(char * hold, int * holdsize,  int (*read_func)(),
              void * stream)
/*
    Calculates the size in bytes of a GRIB product.

    hold = buffer to hold product,
    holdsize = number of bytes of the product in the hold buffer.  Note that
               this increases if necessary as more bytes are read.
    read_func = function used to read the product,
    stream = data describing the input stream (eg. FILE *).

*/
{
int length, numread, num, hsize = *holdsize;
int section2, section3 ;
unsigned char * phold = (unsigned char *) hold;

/* Need more bytes to decide on which version of GRIB */
    num = 24;                             /* put first 24 bytes in buffer */
    if ( hsize < num)
    {
        numread = read_func( hold + hsize, num - hsize, stream);
        if ( numread <= 0 )               /* eg. on EOF */
        {
            *holdsize -= numread;
            return *holdsize;
        }
        hsize = num;
    }
    *holdsize = hsize;

/* See if the GRIB version is 0 ... */
    if ( len3oct(phold+4) == 24 )
    {
       length = 28;                          /* GRIB + section 1 */

/*     Check for presence of sections 2 and 3 */
        section2 = ( hold[11] & BIT1 ) ;
        section3 = ( hold[11] & BIT2 ) ;

/*     Add up all lengths of sections */
        return lentotal(hold, holdsize, length, section2, section3, 
                        read_func, stream);
    }

/* ... or version 1 ... */
    if ( ( hold[21] != 0 ) || ( hold[22] != 0 ) ) 
        return len3oct(phold+4);

/* ... or version -1 ... */
    length = 24;                             /* GRIB + section 1 */

/* Check for presence of sections 2 and 3 */
    section2 = ( hold[7] & BIT8 ) ;
    section3 = ( hold[7] & BIT7 ) ;


/* Add up all lengths of sections */
    return lentotal(hold, holdsize, length, section2, section3, 
                    read_func, stream);

}

static int lentotal(char *hold, int *holdsize, int length, int section2, 
                     int section3, int (*read_func)(), void *stream)
/*
    Returns the total length in bytes of all sections of the GRIB product.

    hold = buffer to hold product,
    holdsize = number of bytes of the product in the hold buffer.  Note that
               this increases if necessary as more bytes are read.
    length = length of (section 0 + section 1).
    section2 is TRUE if section 2 is present in the product.
    section3 is TRUE if section 3 is present in the product.
    read_func = function used to read the product,
    stream = data describing the input stream (eg. FILE *).
*/
{
int sect_5_len = 4;                 /* section 5 length: 7777 */

int numread, hsize = *holdsize;
unsigned char * phold = (unsigned char *) hold;
int next, next_sec = 4;

/* Adjust count of sections to check */
    if ( section2 ) next_sec--;
    if ( section3 ) next_sec--;

/* Get the size of the next section */
    if ( hsize < length)
    {
        numread = read_func( hold + hsize, length - hsize, stream);
        if ( numread <= 0 )               /* eg. on EOF */
        {
            *holdsize -= numread;
            return *holdsize;
        }
        hsize = length;
    } 
    *holdsize = hsize;

/* Get the size of remaining sections */
    for ( next = next_sec; next < 5 ; next++ )
    {
        if ( hsize < (length+4))
        {
            numread = read_func( hold+hsize, (length+4)-hsize, stream);
            if ( numread <= 0 )               /* eg. on EOF */
            {
                *holdsize -= numread;
                return *holdsize;
            }
            hsize = length + 4;
        } 
        *holdsize = hsize;
        length += len3oct(phold+length);
    }

/* Add on the size of section 5 */
    length += sect_5_len;
    if ( hsize < length)
    {
        numread = read_func( hold + hsize, length - hsize, stream);
        if ( numread <= 0 )               /* eg. on EOF */
        {
            *holdsize -= numread;
            return *holdsize;
        }
        hsize = length;
    } 
    *holdsize = hsize;

    return length;
}

static int bufrsize(char * hold, int * holdsize,  int (*read_func)(),
                     void * stream)
/*
    Returns the size in bytes of the BUFR code product.

    hold = buffer to hold product,
    leng = number of bytes of the product in the hold buffer.  Note that
           this increases if necessary as more bytes are read.
    read_func = function used to read the product,
    stream = data describing the input stream (eg. FILE *).
*/
{
unsigned char * phold = (unsigned char *) hold;
int numread, hsize = *holdsize;
int sect_0_len = 4;                 /* section 0 length: BUFR */
int sect_5_len = 4;                 /* section 5 length: 7777 */
int num, length;
int next, next_len, next_sec = 3;


/* Need more bytes to decide on which version of BUFR */
    num = 24;                             /* put first 24 bytes in buffer */
    if ( hsize < num)
    {
        numread = read_func( hold + hsize, num - hsize, stream);
        if ( numread <= 0 )               /* eg. on EOF */
        {
            *holdsize -= numread;
            return *holdsize;
        }
        hsize = num;
    }
    *holdsize = hsize;

/* If it's Edition 2, or later, octets 5-7 give the full product size ... */
    if ( hold[7] > 1 ) return len3oct(phold+4);

/* Otherwise, we have to step through the individual sections
   adding up the lengths */

/* Add on the length of section 1 and make sure enough of product is in 
   memory to continue */
    length = sect_0_len + len3oct(phold+4);
    if ( hsize < (length+4))
    {
        numread = read_func( hold+hsize, (length+4)-hsize, stream);
        if ( numread <= 0 )               /* eg. on EOF */
        {
            *holdsize -= numread;
            return *holdsize;
        }
        hsize = length + 4;
        *holdsize = hsize;
    } 

/* Check for presence of section 2 */
    if ( hold[11] & BIT1 ) next_sec = 2;

/* Get the size of remaining sections */
    for ( next = next_sec; next < 5 ; next++ )
    {
        length += len3oct(phold+length);
        if ( hsize < (length+4))
        {
            numread = read_func( hold+hsize, (length+4)-hsize, stream);
            if ( numread <= 0 )               /* eg. on EOF */
            {
                *holdsize -= numread;
                return *holdsize;
            }
            hsize = length + 4;
            *holdsize = hsize;
        } 
    }

/* Add on the size of section 5 */
    length += sect_5_len;

    return length;

}

/*void GRIBREAD(char * buffer, int * buffsize, int * readsize,
              int * status, FILE ** stream)*/
void GRIBREAD(char * buffer, int * buffsize, int * readsize,
              int * status, int * stream)
/*
    Called as a FORTRAN subroutine:

        CALL GRIBREAD( KARRAY, KINLEN, KOUTLEN, IRET, KUNIT )

*/
{
int holdsize = *buffsize;

/*  Read GRIB product */
    *status =  readprod( "GRIB", buffer, &holdsize, file_read,fptable[*stream-OFFFILES]);
    *readsize = (int) abs( holdsize );
    return;

}

void BUFRREAD(char * buffer, int * buffsize, int * readsize,
              int * status, int * stream)
/*
    Called as a FORTRAN subroutine:

        CALL BUFRREAD( KARRAY, KINLEN, KOUTLEN, IRET, KUNIT )

*/
{
int holdsize = *buffsize;

/*  Read BUFR product */
    *status =  readprod( "BUFR", buffer, &holdsize, file_read,fptable[*stream-OFFFILES]);
    *readsize = (int) abs( holdsize );
    return;

}

static int tide_budg_size(char * hold, int * holdsize,  int (*read_func)(),
                     void * stream)
/*
    Returns the size in bytes of the TIDE code product.

    hold = buffer to hold product,
    leng = number of bytes of the product in the hold buffer.  Note that
           this increases if necessary as more bytes are read.
    read_func = function used to read the product,
    stream = data describing the input stream (eg. FILE *).
*/
{
unsigned char * phold = (unsigned char *) hold;
int numread, hsize = *holdsize;
int sect_0_len = 4;                 /* section 0 length: TIDE */
int sect_5_len = 4;                 /* section 5 length: 7777 */
int num, length;
int next, next_len, next_sec = 3;


/* Need more bytes to get length of section 1 */
    num = 8;                             /* put first 8 bytes in buffer */
    if ( hsize < num)
    {
        numread = read_func( hold + hsize, num - hsize, stream);
        if ( numread <= 0 )               /* eg. on EOF */
        {
            *holdsize -= numread;
            return *holdsize;
        }
        hsize = num;
    }
    *holdsize = hsize;

/* Have to step through the individual sections adding up the lengths */

/* Add on the length of section 1 and make sure enough of product is in 
   memory to continue */
    length = sect_0_len + len3oct(phold+4);
    if ( hsize < (length+4))
    {
        numread = read_func( hold+hsize, (length+4)-hsize, stream);
        if ( numread <= 0 )               /* eg. on EOF */
        {
            *holdsize -= numread;
            return *holdsize;
        }
        hsize = length + 4;
        *holdsize = hsize;
    } 

/* Get the size of remaining section */
    length += len3oct(phold+length);
    if ( hsize < (length+4))
    {
        numread = read_func( hold+hsize, (length+4)-hsize, stream);
        if ( numread <= 0 )               /* eg. on EOF */
        {
            *holdsize -= numread;
            return *holdsize;
        }
        hsize = length + 4;
        *holdsize = hsize;
    } 

/* Add on the size of section 5 */
    length += sect_5_len;

    return length;

}
